// kenlfn.i
// long filename routines for Digital Mars dos32 (X32 dos extender)

#ifndef __klrlfn__
#define __klrlfn__

#define MAXPATH 260
#define MAXSPATH 128

int getshortpath(char *spath, char *lpath);
  // returns 0 on failure, allow 128 bytes for spath
int getlongpath(char *lpath, char *spath);
  // returns 0 on failure, allow 260 bytes for lpath
int lfnrename(char *old, char *new);
  // returns 0 on failure, e.g. if new name already exists
FILE *lfnfopen(char *name, char *mode);
  // returns a FILE pointer
struct lfnFIND *lfnfindfirst(char *filespec);
  // returns NULL if no entry is found
struct lfnFIND *lfnfindnext();
  // returns NULL if no entry is found
void lfnfindclose();

struct lfnFIND
  {
  unsigned char ff_attrib;
  unsigned char ff_attribex[3];
  unsigned int  ff_ctimelo;
  unsigned int  ff_ctimehi;
  unsigned int  ff_atimelo;
  unsigned int  ff_atimehi;
  unsigned int  ff_mtimelo;
  unsigned int  ff_mtimehi;
  unsigned int  ff_sizehi;
  unsigned int  ff_sizelo;
  unsigned char ff_buffer[8];
  unsigned char ff_longname[260];
  unsigned char ff_shortname[14];
  };

//---------------Implementation-------------------
// 4k of dos memory will be permanently reserved, one time only.
// It will be used for two strings and one struct lfnFIND
// at offsets of 0, 1+MAXPATH, and 2+MAXPATH+MAXPATH

#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <io.h>

#define FFBSIZE 318

int lfn_supported = 0;
unsigned short user_dos_seg = 0;
char *user_dos_ptr = NULL;
unsigned short active_find_handle = 0;
struct lfnFIND lfn_ffb;

void find_dos_seg()
  {
  union REGS reg;
  static int btdt = 0;
  if(btdt) return;  // been there, done that
  btdt=1;
  if(sizeof(int)!=4) return;  // failure, not dos32
  reg.x.ax=0x3000;  // get dos version
  int86(0x21,&reg,&reg);
  if(reg.h.al<7) return;  // failure, version < 7
  reg.x.ax=0x4800; // allocate dos memory
  reg.x.bx=0x100;  // 256 paragraphs, 4096 bytes
  int86(0x21,&reg,&reg);
  if(reg.e.cflag) return; // failure, no dos seg available
  lfn_supported=1;  // success
  user_dos_seg=reg.x.ax;
  user_dos_ptr=(char *)reg.e.ebx;
  }

void copy_to_dos_seg(char *str, unsigned short offset, int nbytes)
  {
  char *ptr;
  find_dos_seg();
  if(!lfn_supported) return;
  ptr=offset+user_dos_ptr;
  memcpy(ptr,str,nbytes);
  }

void copy_from_dos_seg(char *str, unsigned short offset, int nbytes)
  {
  char *ptr;
  find_dos_seg();
  if(!lfn_supported) return;
  ptr=offset+user_dos_ptr;
  memcpy(str,ptr,nbytes);
  }

struct lfnFIND *lfnfindfirst(char *filespec)
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return NULL;
  copy_to_dos_seg(filespec,0,1+strlen(filespec));
  reg.x.ax=0x714E;
  reg.x.cx=0x00FF;  // any attribute
  reg.x.si=0x0001;
  sreg.ds=user_dos_seg;  // SEG(filespec);
  reg.x.dx=0;            // OFF(filespec);
  sreg.es=user_dos_seg;        // SEG(ffb);
  reg.x.di=2+MAXPATH+MAXPATH;  // OFF(ffb);
  reg.x.cflag=1;
  int86x_real(0x21,&reg,&reg,&sreg);
  if(reg.x.cflag) return NULL;
  copy_from_dos_seg((char *)&lfn_ffb,2+MAXPATH+MAXPATH,FFBSIZE);
  active_find_handle=reg.x.ax;
  return &lfn_ffb;
  }

struct lfnFIND *lfnfindnext()
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return;
  if(!active_find_handle) return NULL;
  reg.x.ax=0x714F;
  reg.x.bx=active_find_handle;
  reg.x.si=0x0001;
  sreg.es=user_dos_seg;        // SEG(ffb);
  reg.x.di=2+MAXPATH+MAXPATH;  // OFF(ffb);
  reg.x.cflag=1;
  int86x_real(0x21,&reg,&reg,&sreg);
  if(reg.x.cflag) return NULL;
  copy_from_dos_seg((char *)&lfn_ffb,2+MAXPATH+MAXPATH,FFBSIZE);
  return &lfn_ffb;
  }

void lfnfindclose()
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return;
  if(!active_find_handle) return;
  reg.x.ax=0x71A1;
  reg.x.bx=active_find_handle;
  int86x_real(0x21,&reg,&reg,&sreg);
  active_find_handle=0;
  }

int getshortpath(char *spath, char *lpath)
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return;
  spath[0]='\0';
  copy_to_dos_seg(lpath,0,1+strlen(lpath));
  copy_to_dos_seg(spath,1+MAXPATH,1+strlen(spath));
  segread(&sreg);
  reg.x.ax=0x7160;
  reg.x.cx=0x0001;
  sreg.ds=user_dos_seg;  // SEG(lpath);
  reg.x.si=0;            // OFF(lpath);
  sreg.es=user_dos_seg;  // SEG(spath);
  reg.x.di=1+MAXPATH;    // OFF(spath);
  int86x_real(0x21,&reg,&reg,&sreg);
  copy_from_dos_seg(spath,1+MAXPATH,MAXSPATH);
  spath[MAXSPATH-1]='\0';  // for safety
  return (spath[0]=='\0') ? 0 : 1;
  }

int getlongpath(char *lpath, char *spath)
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return;
  lpath[0]='\0';
  copy_to_dos_seg(spath,0,1+strlen(spath));
  copy_to_dos_seg(lpath,1+MAXPATH,1+strlen(lpath));
  segread(&sreg);
  reg.x.ax=0x7160;
  reg.x.cx=0x0002;
  sreg.ds=user_dos_seg;  // SEG(spath);
  reg.x.si=0;            // OFF(spath);
  sreg.es=user_dos_seg;  // SEG(lpath);
  reg.x.di=1+MAXPATH;    // OFF(lpath);
  int86x_real(0x21,&reg,&reg,&sreg);
  copy_from_dos_seg(lpath,1+MAXPATH,MAXPATH);
  lpath[MAXPATH-1]='\0';  // for safety
  return (lpath[0]=='\0') ? 0 : 1;
  }

FILE *lfnfopen(char *name, char *mode)
  {
  char spath[128];
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return fopen(name,mode);
  if((mode[0]=='w')||(mode[0]=='a'))
    {
    copy_to_dos_seg(name,0,1+strlen(name));
    segread(&sreg);
    reg.x.ax=0x716C;
    reg.x.bx=1;    // mode "write"
    reg.x.cx=0;    // attribute "normal"
    reg.x.dx=0x10; // create it unless it already exists
    reg.x.di=1;    // curly hint
    sreg.ds=user_dos_seg;  // SEG(name);
    reg.e.esi=0;           // OFF(name);
    int86x_real(0x21,&reg,&reg,&sreg);
    if(!reg.x.cflag)  // close it
      {
      reg.x.bx=reg.x.ax;
      reg.x.ax=0x3E00;
      int86_real(0x21,&reg,&reg);
      }
    }
  // get it's short pathname
  if(!getshortpath(spath,name)) return NULL;
  // open it using the short pathname
  return fopen(spath,mode);
  }

int lfnrename(char *old, char *new)
  {
  union REGS reg;
  struct SREGS sreg;
  find_dos_seg();
  if(!lfn_supported) return 0;
  copy_to_dos_seg(old,0,1+strlen(old));
  copy_to_dos_seg(new,1+MAXPATH,1+strlen(new));
  segread(&sreg);
  reg.x.ax = 0x7156;
  sreg.ds = user_dos_seg;  // SEG(old);
  reg.x.dx = 0;            // OFF(old);
  sreg.es = user_dos_seg;  // SEG(new);
  reg.x.di = 1+MAXPATH;    // OFF(new);
  reg.x.cflag = 1;
  int86x_real(0x21,&reg,&reg,&sreg);
  return reg.x.cflag ? 0 : 1;
  }

#endif
